/*
 * Copyright (c) 2015 IWAMOTO Kouichi
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "teraterm.h"
#include "tttypes.h"
#include "ttplugin.h"
#include "tt_res.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "compat_w95.h"

#define ORDER 5900
#define SECTION "Color Theme"

#define ID_MENUID_BASE 55201
#define MAX_MENU_ITEMS 20

static HANDLE hInst; /* Instance handle of TTX*.DLL */

typedef struct {
	BOOL change;
	BOOL enable;
	COLORREF fg;
	COLORREF bg;
} TColorSetting;

typedef struct {
	BOOL change;
	BOOL enable;
	COLORREF color[16];
} TAnsiColorSetting;

typedef struct {
	char name[50];
	TColorSetting vt;
	TColorSetting bold;
	TColorSetting blink;
	TColorSetting reverse;
	TColorSetting url;
	TAnsiColorSetting ansicolor;
} TColorTheme;

typedef struct {
	PTTSet ts;
	PComVar cv;
	HMENU SetupMenu;
	HMENU ThemeMenu;
	BOOL ReplaceWinDlg;
	PReadIniFile origReadIniFile;
	int MenuItems;
	TColorTheme DefaultColor;
	TColorTheme ThemeList[MAX_MENU_ITEMS];
} TInstVar;

static TInstVar FAR * pvar;

/* WIN32 allows multiple instances of a DLL */
static TInstVar InstVar;

#define GetFileMenu(menu)    GetSubMenuByChildID(menu, ID_FILE_NEWCONNECTION)
#define GetEditMenu(menu)    GetSubMenuByChildID(menu, ID_EDIT_COPY2)
#define GetSetupMenu(menu)   GetSubMenuByChildID(menu, ID_SETUP_TERMINAL)
#define GetControlMenu(menu) GetSubMenuByChildID(menu, ID_CONTROL_RESETTERMINAL)
#define GetHelpMenu(menu)    GetSubMenuByChildID(menu, ID_HELP_ABOUT)

HMENU GetSubMenuByChildID(HMENU menu, UINT id) {
	int i, j, items, subitems, cur_id;
	HMENU m;

	items = GetMenuItemCount(menu);

	for (i=0; i<items; i++) {
		if (m = GetSubMenu(menu, i)) {
			subitems = GetMenuItemCount(m);
			for (j=0; j<subitems; j++) {
				cur_id = GetMenuItemID(m, j);
				if (cur_id == id) {
					return m;
				}
			}
		}
	}
	return NULL;
}

BOOL GetNthString(PCHAR Source, int Nth, int Size, PCHAR Dest)
{
	int i, j, k;

	i = 1;
	j = k = 0;

	while (i<Nth && Source[j] != 0) {
		if (Source[j++] == ',') {
			i++;
		}
	}

	if (i == Nth) {
		while (Source[j] != 0 && Source[j] != ',' && k<Size-1) {
			Dest[k++] = Source[j++];
		}
	}

	Dest[k] = 0;
	return (i>=Nth);
}

void GetNthNum(PCHAR Source, int Nth, int far *Num)
{
	char T[15];

	GetNthString(Source, Nth, sizeof(T), T);
	if (sscanf(T, "%d", Num) != 1) {
		*Num = 0;
	}
}

void InitMenu() {
	int i;

	if (pvar->ThemeMenu != NULL) {
		DestroyMenu(pvar->ThemeMenu);
	}

	if (pvar->MenuItems > 0) {
		pvar->ThemeMenu = CreateMenu();

		for (i=0; i < pvar->MenuItems; i++) {
			InsertMenu(pvar->ThemeMenu, -1, MF_BYPOSITION, ID_MENUID_BASE+i+1, pvar->ThemeList[i].name);
		}

		InsertMenu(pvar->ThemeMenu, -1, MF_BYPOSITION | MF_SEPARATOR, 0, NULL);
		InsertMenu(pvar->ThemeMenu, -1, MF_BYPOSITION, ID_MENUID_BASE, "Default");
	}
	else {
		pvar->ThemeMenu = NULL;
	}
}

static void PASCAL FAR TTXInit(PTTSet ts, PComVar cv) {
	pvar->ts = ts;
	pvar->cv = cv;
	pvar->ReplaceWinDlg = FALSE;
	pvar->ThemeMenu = NULL;
	pvar->MenuItems = 0;
}

static BOOL FAR PASCAL TTXSetupWin(HWND parent, PTTSet ts) {
	pvar->ReplaceWinDlg = FALSE;
	return (TRUE);
}

static void PASCAL FAR TTXGetUIHooks(TTXUIHooks FAR * hooks) {
	if (pvar->ReplaceWinDlg) {
		*hooks->SetupWin = TTXSetupWin;
	}
	return;
}

void ReadANSIColorSetting(TAnsiColorSetting *color, int num, char *fn) {
	char Key[32], Buff[200];
	int c, r, g, b;

	_snprintf_s(Key, sizeof(Key), _TRUNCATE, "ANSIColor_%d", num + 1);
	GetPrivateProfileString(SECTION, Key, "0", Buff, sizeof(Buff), fn);

	GetNthNum(Buff, 1, &c);
	color->change = c;

	GetNthNum(Buff, 2, &c);
	color->enable = c;

	for (c=0; c<16; c++) {
		GetNthNum(Buff, c * 3 + 3, &r);
		GetNthNum(Buff, c * 3 + 4, &g);
		GetNthNum(Buff, c * 3 + 5, &b);
		color->color[c] = RGB(r, g, b);
	}

	return;
}

void ReadColorSetting(TColorSetting *color, char *ent, int num, char *fn) {
	char Key[32], Buff[200];
	int c, r, g, b;

	_snprintf_s(Key, sizeof(Key), _TRUNCATE, "%s_%d", ent, num + 1);
	GetPrivateProfileString(SECTION, Key, "0", Buff, sizeof(Buff), fn);

	GetNthNum(Buff, 1, &c);
	color->change = c;

	GetNthNum(Buff, 2, &c);
	color->enable = c;

	GetNthNum(Buff, 3, &r);
	GetNthNum(Buff, 4, &g);
	GetNthNum(Buff, 5, &b);
	color->fg = RGB(r, g, b);

	GetNthNum(Buff, 6, &r);
	GetNthNum(Buff, 7, &g);
	GetNthNum(Buff, 8, &b);
	color->bg = RGB(r, g, b);

	return;
}

void GetDefaultColor(TColorSetting *tc, COLORREF *color, int field, PTTSet ts) {
	tc->change = TRUE;
	if (field)
		tc->enable = (ts->ColorFlag & field) != 0;
	else // VTColor
		tc->enable = TRUE;
	tc->fg = color[0];
	tc->bg = color[1];

	return;
}

static void PASCAL FAR ColorThemeReadIniFile(PCHAR fn, PTTSet ts) {
	int i;
	char Key[20];

	(pvar->origReadIniFile)(fn, ts);

	GetDefaultColor(&(pvar->DefaultColor.vt), ts->VTColor, 0, ts);
	GetDefaultColor(&(pvar->DefaultColor.bold), ts->VTBoldColor, CF_BOLDCOLOR, ts);
	GetDefaultColor(&(pvar->DefaultColor.blink), ts->VTBlinkColor, CF_BLINKCOLOR, ts);
	GetDefaultColor(&(pvar->DefaultColor.reverse), ts->VTReverseColor, CF_REVERSECOLOR, ts);
	GetDefaultColor(&(pvar->DefaultColor.url), ts->URLColor, CF_URLCOLOR, ts);

	pvar->DefaultColor.ansicolor.change = 1;
	pvar->DefaultColor.ansicolor.enable = (ts->ColorFlag & CF_ANSICOLOR) != 0;
	for (i=0; i<16; i++) {
		pvar->DefaultColor.ansicolor.color[i] = ts->ANSIColor[i];
	}

	for (i=0; i<MAX_MENU_ITEMS; i++) {
		_snprintf_s(Key, sizeof(Key), _TRUNCATE, "Theme_%d", i+1);
		GetPrivateProfileString(SECTION, Key, "", pvar->ThemeList[i].name,
		    sizeof(pvar->ThemeList[i].name), fn);
		if (pvar->ThemeList[i].name[0] == '\0')
			break;

		ReadColorSetting(&(pvar->ThemeList[i].vt), "VTColor", i, fn);
		ReadColorSetting(&(pvar->ThemeList[i].bold), "BoldColor", i, fn);
		ReadColorSetting(&(pvar->ThemeList[i].blink), "BlinkColor", i, fn);
		ReadColorSetting(&(pvar->ThemeList[i].reverse), "ReverseColor", i, fn);
		ReadColorSetting(&(pvar->ThemeList[i].url), "URLColor", i, fn);

		ReadANSIColorSetting(&(pvar->ThemeList[i].ansicolor), i, fn);
	}

	pvar->MenuItems = i;
}

static void PASCAL FAR TTXGetSetupHooks(TTXSetupHooks FAR * hooks) {
	pvar->origReadIniFile = *hooks->ReadIniFile;
	*hooks->ReadIniFile = ColorThemeReadIniFile;
}

static void PASCAL FAR TTXModifyMenu(HMENU menu) {
	MENUITEMINFO mi;

	pvar->SetupMenu = GetSetupMenu(menu);

	if (pvar->MenuItems > 0) {
		InitMenu();

		memset(&mi, 0, sizeof(mi));
		mi.cbSize = sizeof(mi);
		mi.fMask  = MIIM_TYPE | MIIM_SUBMENU;
		mi.fType  = MFT_STRING;
		mi.hSubMenu = pvar->ThemeMenu;
		mi.dwTypeData = "Color The&me";
		InsertMenuItem(pvar->SetupMenu, ID_SETUP_ADDITIONALSETTINGS, FALSE, &mi);
	}
}

/*
static void PASCAL FAR TTXModifyPopupMenu(HMENU menu) {
	if (menu == pvar->ResizeMenu) {
		UpdateMenu();
	}
}
*/

static int PASCAL FAR TTXProcessCommand(HWND HWin, WORD cmd) {
	int num, i;
	TColorTheme *theme;
	if (ID_MENUID_BASE <= cmd && cmd <= ID_MENUID_BASE + pvar->MenuItems) {
		num = cmd - ID_MENUID_BASE;
		pvar->ReplaceWinDlg = TRUE;

		if (num == 0) {
			theme = &(pvar->DefaultColor);
		}
		else {
			theme = &(pvar->ThemeList[num-1]);
		}
		if (theme->vt.change) {
			pvar->ts->VTColor[0] = theme->vt.fg;
			pvar->ts->VTColor[1] = theme->vt.bg;
		}

		if (theme->bold.change) {
			if (theme->bold.enable) {
				pvar->ts->ColorFlag |= CF_BOLDCOLOR;
				pvar->ts->VTBoldColor[0] = theme->bold.fg;
				pvar->ts->VTBoldColor[1] = theme->bold.bg;
			}
			else {
				pvar->ts->ColorFlag &= ~CF_BOLDCOLOR;
			}
		}

		if (theme->blink.change) {
			if (theme->blink.enable) {
				pvar->ts->ColorFlag |= CF_BLINKCOLOR;
				pvar->ts->VTBlinkColor[0] = theme->blink.fg;
				pvar->ts->VTBlinkColor[1] = theme->blink.bg;
			}
			else {
				pvar->ts->ColorFlag &= ~CF_BLINKCOLOR;
			}
		}

		if (theme->reverse.change) {
			if (theme->reverse.enable) {
				pvar->ts->ColorFlag |= CF_REVERSECOLOR;
				pvar->ts->VTReverseColor[0] = theme->reverse.fg;
				pvar->ts->VTReverseColor[1] = theme->reverse.bg;
			}
			else {
				pvar->ts->ColorFlag &= ~CF_REVERSECOLOR;
			}
		}

		if (theme->url.change) {
			if (theme->url.enable) {
				pvar->ts->ColorFlag |= CF_URLCOLOR;
				pvar->ts->URLColor[0] = theme->url.fg;
				pvar->ts->URLColor[1] = theme->url.bg;
			}
			else {
				pvar->ts->ColorFlag &= ~CF_URLCOLOR;
			}
		}

		if (theme->ansicolor.change) {
			for (i=0; i<16; i++) {
				pvar->ts->ANSIColor[i] = theme->ansicolor.color[i];
			}
		}

		// Call Setup-Window dialog
		SendMessage(HWin, WM_COMMAND, MAKELONG(ID_SETUP_WINDOW, 0), 0);
		return 1;
	}
	return 0;
}

static TTXExports Exports = {
  sizeof(TTXExports),
  ORDER,

  TTXInit,
  TTXGetUIHooks,
  TTXGetSetupHooks,
  NULL, // TTXOpenTCP,
  NULL, // TTXCloseTCP,
  NULL, // TTXSetWinSize,
  TTXModifyMenu,
  NULL, // TTXModifyPopupMenu,
  TTXProcessCommand,
  NULL, // TTXEnd
};

BOOL __declspec(dllexport) PASCAL FAR TTXBind(WORD Version, TTXExports FAR * exports) {
  int size = sizeof(Exports) - sizeof(exports->size);

  if (size > exports->size) {
    size = exports->size;
  }
  memcpy((char FAR *)exports + sizeof(exports->size),
         (char FAR *)&Exports + sizeof(exports->size),
         size);
  return TRUE;
}

BOOL WINAPI DllMain(HANDLE hInstance, ULONG ul_reason, LPVOID lpReserved)
{
  switch (ul_reason) {
    case DLL_THREAD_ATTACH:
      /* do thread initialization */
      break;
    case DLL_THREAD_DETACH:
      /* do thread cleanup */
      break;
    case DLL_PROCESS_ATTACH:
      /* do process initialization */
      DoCover_IsDebuggerPresent();
      hInst = hInstance;
      pvar = &InstVar;
      break;
    case DLL_PROCESS_DETACH:
      /* do process cleanup */
      break;
  }
  return TRUE;
}
